def get_numeric_ratio(s):
	"""
	gets the ratio of numeric characters to the total alphanumerics
	:type s: str
	:rtype: float or NoneType
	"""
	num_numeric = 0
	num_alphabetic = 0
	if len(s) == 0:
		return None
	else:
		for character in s:
			if character.isnumeric():
				num_numeric += 1
			elif character.isalpha():
				num_alphabetic += 1
		return num_numeric / (num_numeric + num_alphabetic)



from .is_break import is_break


def get_children(tag):
	"""
	:type tag: Tag or list or tuple
	:rtype: NoneType or list[Tag]
	"""
	if isinstance(tag, (list, tuple)):
		return list(tag)

	if not hasattr(tag, 'children'):
		return None
	else:
		children = [x for x in tag.children if x is not None and x != []]

		# remove repeated breaks
		new_children = []
		for child in children:
			if len(new_children) == 0:
				new_children.append(child)
			elif is_break(child) and is_break(new_children[-1]):  # ignore repeated break
				continue
			else:
				new_children.append(child)
		children = new_children

		if len(children) > 0:
			return children
		else:
			return None



import re


def get_font_size(tag):
	try:
		style = tag.attrs.get('style')
		font_size = re.search(r'(?is)(font-size:)(.*?)(px)', str(style)).group(2)
		return float(font_size)
	except AttributeError:
		return None






import re


def has_border(tag):
	try:
		style = tag.attrs.get('style')
		re.search(r'(?is)(border:)', str(style)).group(1)
	except AttributeError:
		return None


def get_invisible_border_size(tag):
	try:
		style = tag.attrs.get('style')
		border_size = re.search(r'(?is)(border:\s*textbox)(.*?)(px)', str(style)).group(2)
		return float(border_size)
	except AttributeError:
		return None


def get_border_size(tag):
	try:
		style = tag.attrs.get('style')
		border_size = re.search(r'(?is)(border:)(\s*\w+\s*)(.*?)(px)', str(style)).group(3)
		return float(border_size)
	except AttributeError:
		return None
	except ValueError as e:
		print(style)
		raise e


def get_visible_border_size(tag):
	if has_border(tag):
		if get_invisible_border_size(tag):
			return None
		else:
			return get_border_size(tag)
	else:
		return None


from .get_children import get_children
from .is_break import is_break


# if a tag starts with a br this might belong to the previous tag which would make that one a title
def previous_tag_might_be_title(tag, children=None):
	if children is None:
		tag_children = get_children(tag)
	else:
		tag_children = children

	if tag_children and len(tag_children) > 1 and is_break(tag_children[0]):
		return True
	else:
		return False



import os
from disk import Path
import pdfminer
import pdfminer.layout
import pdfminer.high_level
import sys
from time import sleep
from .exceptions import PdfError

OUTPUT_TYPES = (
    (".htm", "html"),
    (".html", "html"),
    (".xml", "xml"),
    (".tag", "tag")
)


def convert_pdf_to_html_using_command(pdf_path, html_path):
    if isinstance(pdf_path, Path):
        pdf_path = pdf_path.path
    if isinstance(html_path, Path):
        html_path = html_path.path
    os.system(f'pdf2txt.py -o "{html_path}" "{pdf_path}"')


def extract_text(
        files=None, outfile='-',
        no_laparams=False, all_texts=None, detect_vertical=None,
        word_margin=None, char_margin=None, line_margin=None,
        boxes_flow=None, output_type='text', codec='utf-8',
        strip_control=False, maxpages=0, page_numbers=None,
        password="", scale=1.0, rotation=0, layoutmode='normal',
        output_dir=None, debug=False, disable_caching=False,
        **kwargs
):
    files = files or []
    if not files:
        raise ValueError("Must provide files to work upon!")

    # If any LAParams group arguments were passed,
    # create an LAParams object and
    # populate with given args. Otherwise, set it to None.
    if not no_laparams:
        laparams = pdfminer.layout.LAParams()
        params = ("all_texts", "detect_vertical", "word_margin", "char_margin", "line_margin", "boxes_flow")
        for param in params:
            paramv = locals().get(param, None)
            if paramv is not None:
                setattr(laparams, param, paramv)
    else:
        laparams = None

    if output_type == "text" and outfile != "-":
        for override, alttype in OUTPUT_TYPES:
            if outfile.endswith(override):
                output_type = alttype

    if outfile == "-":
        outfp = sys.stdout
        if outfp.encoding is not None:
            codec = 'utf-8'

        for fname in files:
            with open(fname, "rb") as fp:
                pdfminer.high_level.extract_text_to_fp(fp, **locals())
        return outfp

    else:
        with open(outfile, "wb") as outfp:
            for fname in files:
                with open(fname, "rb") as fp:
                    pdfminer.high_level.extract_text_to_fp(fp, **locals())
        return None


def convert_pdf_to_html(pdf_path, html_path, num_tries=3):
    if isinstance(pdf_path, Path):
        pdf_path = pdf_path.path
    if isinstance(html_path, Path):
        html_path = html_path.path

    for i in range(num_tries):
        try:
            extract_text(
                files=[pdf_path], debug=False, disable_caching=False, page_numbers=None, pagenos=None, maxpages=0,
                password='', rotation=0, no_laparams=False, detect_vertical=False, char_margin=2.0,
                word_margin=0.1, line_margin=0.5, boxes_flow=0.5, all_texts=True, outfile=html_path,
                output_type='html', codec='utf-8',
                output_dir=None, layoutmode='normal', scale=1.0, strip_control=False
            )
            if Path(html_path).exists():
                break
            else:
                sleep(2 ** i)
        except Exception as e:
            if i < 2:
                continue
            else:
                raise PdfError(f'error converting "{pdf_path}" to html: {e}')
    else:
        error_message = f'error converting "{pdf_path}" to html'
        Path(html_path).save(obj=error_message)
        raise PdfError(error_message)
    return html_path



from nltk import word_tokenize


def get_words(string, language='english'):
	return word_tokenize(string, language=language)


def count_words(string, language='english'):
	return len(get_words(string=string, language=language))


def count_english_words(x, vocabulary, case_sensitive=False):
	if isinstance(x, (list, tuple)):
		words = x
	elif isinstance(x, str):
		words = get_words(string=x, language='english')

	if hasattr(vocabulary, 'words'):
		vocabulary = vocabulary.words()
	elif isinstance(vocabulary, (list, tuple, set)):
		pass
	else:
		raise TypeError(f'{vocabulary} is of type {type(vocabulary)}!')

	num_english_words = 0
	for word in words:
		if len(word) < 2:
			if word in ['a', 'A', 'I']:
				num_english_words += 1
			elif not case_sensitive and word == 'i':
				num_english_words += 1
		else:
			if word in vocabulary:
				num_english_words += 1
			elif not case_sensitive:
				if word.capitalize() in vocabulary or word.lower() in vocabulary or word.upper() in vocabulary:
					num_english_words += 1
	return num_english_words


def get_ratio_of_english_words(x, vocabulary=None, case_sensitive=False):
	if isinstance(x, (list, tuple)):
		words = x
	elif isinstance(x, str):
		words = get_words(string=x, language='english')

	if isinstance(x, (list, tuple)):
		words = x
	elif isinstance(x, str):
		words = get_words(string=x, language='english')

	num_english_words = count_english_words(x=words, vocabulary=vocabulary, case_sensitive=case_sensitive)

	return num_english_words / len(words)



from disk import Path
from silverware import HtmlFile, Spoon
from silverware.htmltext import HtmlText
from chronometry import get_elapsed
from chronometry import get_now

import re

from .convert_pdf_to_html import convert_pdf_to_html
from .exceptions import ConversionError


class Pdf(Path):
	def __init__(self, string, vocabulary=None, show_size=False, html_directory=None, num_jobs=1):
		super().__init__(string=string, show_size=show_size)
		self._html_file = None
		self._spoon = None
		self._paragraphs = None
		self._vocabulary = vocabulary

		if html_directory is not None:
			html_directory = Path(html_directory)
		self._html_directory = html_directory

		self._elapsed = {}
		self._html_text = None
		self._num_jobs = num_jobs

	@property
	def html_directory(self):
		if self._html_directory is None:
			self._html_directory = self.parent_directory + (self.name + '_helpers')
		return self._html_directory

	@property
	def html_path(self):
		return self.html_directory + (self.name + '.html')

	def convert_to_html(self, ignore_if_exists=True, num_tries=3):
		"""
		:rtype: HtmlFile
		"""
		path = self.html_path

		start_time = get_now()
		for try_num in range(num_tries + 1):
			if path.exists() and ignore_if_exists and path.get_size_kb() > 6:
				break
			else:
				if not path.parent_directory.exists():
					path.parent_directory.make_dir()
				convert_pdf_to_html(pdf_path=self, html_path=path)
				if path.exists() and path.get_size_kb() > 6:
					self._elapsed['html_conversion'] = get_elapsed(start=start_time, unit='s')
					break
		else:
			raise ConversionError(f'could not convert {self.name_and_extension} to html after {num_tries} tries')

		html_file = HtmlFile(path)
		return html_file

	@property
	def html_file(self):
		"""
		:rtype: HtmlFile
		"""
		if self._html_file is None:
			self._html_file = self.convert_to_html()
		return self._html_file

	@property
	def text_path(self):
		"""
		:rtype: Path
		"""
		return self.html_directory + (self.name + '_html_text.pickle')

	@property
	def spoon(self):
		"""
		:rtype: Spoon
		"""
		if self._spoon is None:
			start_time = get_now()
			self._spoon = self.html_file.spoon
			self._elapsed['html_parsing'] = get_elapsed(start=start_time, unit='s')
		return self._spoon

	@property
	def text(self):
		"""
		:rtype: HtmlText
		"""
		if self._html_text is None:
			self._html_text = self.generate_text(ignore_if_exists=True)
		return self._html_text

	def generate_text(self, ignore_if_exists=True):
		"""
		:rtype: HtmlText
		"""
		if self.text_path.exists() and ignore_if_exists:
			text = self.text_path.load()

		else:
			text = HtmlText(
				obj=self,
				ignore_function=lambda x: not re.match(r'^page \d+$', 'Page 02', flags=re.IGNORECASE),
				log=False,
				num_jobs=self._num_jobs
			)
			self.text_path.save(text)

		return text

	def delete_extra_files(self):
		if self.html_directory.exists():
			self.html_directory.delete()



def get_alphabetic_ratio(string):
	"""
	gives the ratio of alphabetic characters to the rest
	:type string: str
	:rtype: float or NoneType
	"""
	length = len(string)
	if length == 0:
		return None
	else:
		return sum([1 for character in string if character.isalpha()]) / length



CAP_EXEMPT = [
	'', 'and', 'as', 'as if', 'as long as', 'at', 'but', 'by', 'even if', 'for', 'from', 'if', 'if only', 'in',
	'into', 'like', 'near', 'now that', 'nor', 'of', 'off', 'on', 'on top of', 'once', 'onto', 'or', 'out of',
	'over', 'past', 'so', 'so that', 'than', 'that', 'till', 'to', 'up', 'upon', 'with', 'when', 'yet'
]


def is_capitalized(string):
	words = string.split()
	for word in words:
		if word not in CAP_EXEMPT and word[0].islower():
			return False
	return True



from .get_children import get_children
from .get_font_size import get_font_size
from .get_border import get_border_size
from .get_border import get_visible_border_size
from .get_raw_text import get_raw_text
from .get_text import get_text
from .is_break import is_break
from .is_tag_title import is_tag_title
from .previous_tag_might_be_title import previous_tag_might_be_title



from .ends_with_punctuation import ends_with_punctuation
from .get_alphabetic import get_alphabetic
from .get_words import count_words
from .get_alphabetic_ratio import get_alphabetic_ratio
from .get_numeric_ratio import get_numeric_ratio
from .is_capitalized import is_capitalized
from .is_short_text import is_short_text
from .strip_non_alphanumeric_from_beginning import strip_non_alphanumeric_from_beginning
from .get_words import get_ratio_of_english_words
from .get_words import get_words
from .get_words import count_words
from .get_words import count_english_words



class PdfError(TypeError):
	pass


class HtmlTextError(RuntimeError):
	pass


class EmptyTagError(HtmlTextError):
	pass


class MissingTagError(HtmlTextError):
	pass


class MissingIndexError(HtmlTextError):
	pass


class AbsorptionError(HtmlTextError):
	pass


class HtmlTextDisplayError(HtmlTextError):
	pass


class ConversionError(PdfError):
	pass



from setuptools import setup, find_packages


try:
	import nltk

	nltk.download('punkt')
	nltk.download('words')

except:
	try:
		import os

		os.system('pip install nltk')

		import nltk

		nltk.download('punkt')
		nltk.download('words')
	except:
		pass


def readme():
	with open('./README.md') as f:
		return f.read()


setup(
	name='pdf',
	version='2020.9.17.1',
	license='MIT',
	author='Idin',
	author_email='py@idin.ca',
	url='https://github.com/idin/pdf',
	keywords='PDF',
	description='Python library for parsing PDFs',
	long_description=readme(),
	long_description_content_type='text/markdown',
	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Developers',
		'Programming Language :: Python :: 3 :: Only',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Topic :: Software Development :: Libraries :: Python Modules'
	],
	packages=find_packages(exclude=("jupyter_tests", ".idea", ".git")),
	install_requires=['pdftotree', 'bs4', 'silverware', 'nltk', 'disk', 'pandas', 'chronometry'],
	python_requires='~=3.6',
	zip_safe=True,
)



from .get_raw_text import get_raw_text


def is_break(tag):
	if tag.name == 'br':
		return True
	else:
		text = get_raw_text(tag)
		if text:
			if '\n' in text and text.strip() == '':
				return True
			else:
				return False
		else:
			return False



from .get_raw_text import get_raw_text
from ..string import get_alphabetic
import re


def get_text(tag, remove_non_alphabetic=True):
	"""
	:type tag: Tag
	:rtype: NoneType or str
	"""
	text = get_raw_text(tag)
	text = text.strip()

	if re.search('^Page \d+$', text):
		return None

	if remove_non_alphabetic:
		alphabetic = get_alphabetic(text)
		if len(alphabetic) < 2:
			return None
		else:
			return text
	else:
		return text



from .get_children import get_children
from .get_text import get_text
from .is_break import is_break
from ..string import ends_with_punctuation
from ..string import is_capitalized
from ..string import count_words


def is_tag_title(tag, children=None):
	if children is None:
		tag_children = get_children(tag)
	else:
		tag_children = children

	# if this tag has only two children and the second one is a br then it's a title
	if len(tag_children) == 2 and is_break(tag_children[-1]):
		tag_children.pop()
		only_child = tag_children[0]
		child_text = get_text(only_child)
		if child_text:
			if not ends_with_punctuation(child_text) and count_words(child_text) < 13:
				return True
			else:
				return is_capitalized(child_text)
		else:
			return False

	else:
		return False
